import contextlib
import socket
import subprocess
import sys
import time

import httpx

from ragna._utils import timeout_after


@contextlib.contextmanager
def background_subprocess(*args, stdout=sys.stdout, stderr=sys.stdout, **kwargs):
    process = subprocess.Popen(*args, stdout=stdout, stderr=stderr, **kwargs)
    try:
        yield process
    finally:
        process.kill()
        process.communicate()


def get_available_port():
    with socket.socket() as s:
        s.bind(("", 0))
        return s.getsockname()[1]


@contextlib.contextmanager
def ragna_api(config):
    config_path = config.local_cache_root / "ragna.toml"
    if not config_path.exists():
        config.to_file(config_path)

    cmd = [sys.executable, "-m", "ragna", "api", "--config", str(config_path)]

    with background_subprocess(cmd):

        @timeout_after(message="Unable to start ragna api")
        def wait_for_ragna_api(poll=0.1):
            url = config.api.url
            while True:
                with contextlib.suppress(httpx.ConnectError):
                    response = httpx.get(url)
                    if response.is_success:
                        return url

                time.sleep(poll)

        yield wait_for_ragna_api()



import abc
import os
from typing import AsyncIterator

import httpx

import ragna
from ragna.core import Assistant, EnvVarRequirement, Requirement, Source


class ApiAssistant(Assistant):
    _API_KEY_ENV_VAR: str

    @classmethod
    def requirements(cls) -> list[Requirement]:
        return [EnvVarRequirement(cls._API_KEY_ENV_VAR)]

    def __init__(self) -> None:
        self._client = httpx.AsyncClient(
            headers={"User-Agent": f"{ragna.__version__}/{self}"},
            timeout=60,
        )
        self._api_key = os.environ[self._API_KEY_ENV_VAR]

    async def answer(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int = 256
    ) -> AsyncIterator[str]:
        async for chunk in self._call_api(  # type: ignore[attr-defined, misc]
            prompt, sources, max_new_tokens=max_new_tokens
        ):
            yield chunk

    @abc.abstractmethod
    async def _call_api(
        self, prompt: str, sources: list[Source], *, max_new_tokens: int
    ) -> AsyncIterator[str]:
        ...



import os

import pytest

from ragna.deploy import Config


@pytest.mark.xfail()
def test_explicit_gt_env_var(mocker, tmp_path):
    explicit = tmp_path / "explicit"

    env_var = tmp_path / "env_var"
    mocker.patch.dict(os.environ, values={"RAGNA_LOCAL_CACHE_ROOT": str(env_var)})

    config = Config(local_cache_root=explicit)

    assert config.local_cache_root == explicit


def test_env_var_gt_config_file(mocker, tmp_path):
    config_file = tmp_path / "config_file"
    config = Config(local_cache_root=config_file)
    assert config.local_cache_root == config_file

    config_path = tmp_path / "ragna.toml"
    config.to_file(config_path)

    env_var = tmp_path / "env_var"
    mocker.patch.dict(os.environ, values={"RAGNA_LOCAL_CACHE_ROOT": str(env_var)})

    config = Config.from_file(config_path)

    assert config.local_cache_root == env_var
